#include <stdio.h>
#include <stdint.h>
#include <stdlib.h>
#include <string.h>

#include "loadpict2.xpm"

static uint8_t lpvmo[22 * 23];
static uint8_t lpclo[22 * 23];
static uint8_t lpcho[2][8 * 256];
static uint8_t cli2o[256];
static uint8_t clo2i[16];
static int ysplit, nch0, nch1;

/*
COLOR_BLACK = 0
COLOR_WHITE = 1
COLOR_RED = 2
COLOR_CYAN = 3
COLOR_PURPLE = 4
COLOR_GREEN = 5
COLOR_BLUE = 6
COLOR_YELLOW = 7                   0  3  5  2  2  7  1
*//*                               0  1  2  3  4  5  6 */
static const uint8_t clmap[16] = { 0, 5, 5, 2, 2, 7, 1, 0, 0, 0, 0, 0, 0, 0, 0 };

static void binprint(const uint8_t *p)
{
    int i;
    uint8_t b;
    for (i = 0; i < 8; ++i) {
        b = *p++;
        fprintf(stdout, "!by %%%c%c%c%c%c%c%c%c\n",
            (b & 0x80) ? '#' : '.',
            (b & 0x40) ? '#' : '.',
            (b & 0x20) ? '#' : '.',
            (b & 0x10) ? '#' : '.',
            (b & 0x08) ? '#' : '.',
            (b & 0x04) ? '#' : '.',
            (b & 0x02) ? '#' : '.',
            (b & 0x01) ? '#' : '.'
        );
    }
}

static int parsenum(char **p_ptr)
{
    char *p = *p_ptr;
    unsigned long int l = strtoul(p, p_ptr, 10);
    if (p == *p_ptr) {
        return -1;
    }
    return (int)(l & 0xffff);
}

static int process(void)
{
    int choi, i, j, ybase, xc, yc, chset;
    uint8_t c, cempty;

    /* check x, y, num of colors */
    {
        int fx, fy, fc, fn;
        char *p = loadpict2[0];
        if ((fx = parsenum(&p)) != (22 * 8)) {
            fprintf(stderr, "invalid x %i\n", fx);
            return 1;
        }
        if ((fy = parsenum(&p)) != (23 * 8)) {
            fprintf(stderr, "invalid y %i\n", fy);
            return 1;
        }
        fc = parsenum(&p);
        if ((fc < 1) || (fc > 16)) {
            fprintf(stderr, "invalid colors %i\n", fc);
            return 1;
        }
        ybase = fc + 1;
        if ((fn = parsenum(&p)) != 1) {
            fprintf(stderr, "invalid n %i\n", fn);
            return 1;
        }
        /* match chars to colors */
        for (i = 0; i < fc; ++i) {
            c = (uint8_t)(loadpict2[i + 1][0]);
            if (i == 0) cempty = c;
            cli2o[c] = clmap[i];
            clo2i[clmap[i]] = c;
            fprintf(stderr, "i %i c %c (%02x) cli2o %i\n", i, c, c, cli2o[c]);
        }
    }

    ysplit = -1;
    chset = 0;
split_found:
    choi = 0;
    nch0 = 0;
    nch1 = 0;

    for (yc = 0; yc < 23; ++yc) {
        int xo, yo, c1, ct, x, y;
        uint8_t bm[8], b;
        if (yc == ysplit) {
            ++chset;
            nch0 = choi;
            choi = 0xff;
        }
        yo = yc * 8;
        for (xc = 0; xc < 22; ++xc) {
            xo = xc * 8;
            c1 = 0;
            for (y = 0; y < 8; ++y) {
                const char *p;
                p = &loadpict2[yo + ybase + y][xo];
                b = 0;
                for (x = 0; x < 8; ++x) {
                    b <<= 1;
                    c = (uint8_t)(p[x]);
                    if (c != cempty) {
                        b |= 1;
                        ct = cli2o[c];
                        if (c1 == 0) {
                            c1 = ct;
                        } else if (c1 != ct) {
                            fprintf(stderr, "color clash at %ix%i (char %ix%i) c1 %i ct %i\n", xo+x, yo+y, xc, yc, c1, ct);
                            return 1;
                        }
                    }
                }
                bm[y] = b;
                lpclo[yc * 22 + xc] = c1;
            }
            /* seach for duplicates in ch */
            for (j = ((chset == 0) ? 0 : (choi + 1)); j < ((chset == 0) ? choi : 0x100); ++j) {
                if (memcmp(&lpcho[chset][j * 8], &bm[0], 8) == 0) {
/*fprintf(stderr, "match with %i (%i, %ix%i)\n", j, choi, xc, yc);*/
                    c = j;
                    goto found_match;
                }
            }
            memcpy(&lpcho[chset][choi * 8], &bm[0], 8);
            c = choi;
            choi += chset ? -1 : 1;
            if ((chset == 0) && (choi == 0xc0)) {
                fprintf(stderr, "y split at %ix%i\n", xc, yc);
                ysplit = yc;
                goto split_found;
            }

            if (chset == 0) {
                ++nch0;
            } else {
                ++nch1;
            }
found_match:
            lpvmo[yc * 22 + xc] = c;
        }
    }

    fprintf(stderr, "chars %i, %i split %i\n", nch0, nch1, ysplit);
    return 0;
}

static void printfont(int chset, int n)
{
    int i;
    fprintf(stdout, "font%i: ; %i chars\n", chset, n);
    for (i = ((chset == 0) ? 0 : (0x100 - n)); i < ((chset == 0) ? n : 0x100); ++i) {
        fprintf(stdout, "; $%02x %i\n", i, chset);
        binprint(&lpcho[chset][i * 8]);
    }
    fprintf(stdout, "font%iend = *\n", chset);
}

static int printpic(void)
{
#if 0
    int i, j, k, chset = 0;
    uint8_t c1;

    for (j = 0; j < 23; ++j) {
        fprintf(stdout, "\"");
        if ((chset == 0) && (j == ysplit)) {
            ++chset;
        }
        for (i = 0; i < 22; ++i) {
            c1 = lpclo[j * 22 + i];
            c1 = clo2i[c1];
            fprintf(stdout, "%02x ", c1);
        }
        fprintf(stdout, "\",\n");
    }
#endif
#if 0
    int i, j, k, chset = 0;
    uint8_t b, c1, c0 = clo2i[0];

    for (j = 0; j < (23 * 8); ++j) {
        fprintf(stdout, "\"");
        if ((chset == 0) && ((j / 8) == ysplit)) {
            ++chset;
        }
        for (i = 0; i < 22; ++i) {
            c1 = lpclo[(j / 8) * 22 + i];
            c1 = clo2i[c1];
            b = lpvmo[(j / 8) * 22 + i];
            b = lpcho[chset][((uint16_t)(b) * 8) + (j & 7)];
            for (k = 0; k < 8; ++k) {
                fprintf(stdout, "%c", (b & (1 << (7 - k))) ? c1 : c0);
            }
        }
        fprintf(stdout, "\",\n");
    }
#endif
#if 1
    int i, j;
    uint8_t c;

    fprintf(stdout, "; autogenerated\nYSPLIT = %i\n* = $1800\n", ysplit > 0 ? ysplit : 0);
    printfont(0, nch0);
    if (ysplit > 0) {
        fprintf(stdout, "* = $1800 - %i * 8\n", nch1);
        printfont(1, nch1);
    }

    fprintf(stdout, "* = $1800 - %i * 8 - 23 * 22 / 2\ncolordata: ; c0 | (c1 << 4)\n", nch1);
    for (j = 0; j < 23; ++j) {
        fprintf(stdout, "!by ");
        for (i = 0; i < 22 / 2; ++i) {
            c = lpclo[j * 22 + i * 2 + 0];
            c |= lpclo[j * 22 + i * 2 + 1] << 4;
            fprintf(stdout, "$%02x", c);
            if (i < (22 / 2 - 1)) {
                fprintf(stdout, ", ");
            }
        }
        fprintf(stdout, "\n");
    }

    fprintf(stdout, "colordataend = *\n!if (* > $1e00) {\n!error \"cd over vm: \", * - $1e00\n}\n* = $1e00\nvmatrix:\n");
    for (j = 0; j < 23; ++j) {
        fprintf(stdout, "!by ");
        for (i = 0; i < 22; ++i) {
            c = lpvmo[j * 22 + i];
            fprintf(stdout, "$%02x", c);
            if (i < (22 - 1)) {
                fprintf(stdout, ", ");
            }
        }
        fprintf(stdout, "\n");
    }

#endif
    return 0;
}

int main(int argc, char **argv)
{
    if (process()) {
        return 1;
    }
    return printpic();
}
